from hwtypes import BitVector
from magma.array import Array
from magma.bit import Bit
from magma.bitutils import clog2, seq2int
from magma.circuit import coreir_port_mapping
from magma.generator import Generator2
from magma.interface import IO
from magma.protocol_type import MagmaProtocol
from magma.t import Type, In, Out
from magma.tuple import Product


class CoreIRCommonLibMuxN(Generator2):
    def __init__(self, N: int, width: int):
        self.name = f"coreir_commonlib_mux{N}x{width}"
        FlatT = Array[width, Bit]
        MuxInT = Product.from_fields("anon", dict(data=Array[N, FlatT],
                                                  sel=Array[clog2(N), Bit]))
        self.io = IO(I=In(MuxInT), O=Out(Array[width, Bit]))
        self.renamed_ports = coreir_port_mapping
        self.coreir_name = "muxn"
        self.coreir_lib = "commonlib"
        self.coreir_genargs = {"width": width, "N": N}
        self.primitive = True
        self.stateful = False

        def simulate(self, value_store, state_store):
            sel = BitVector[clog2(N)](value_store.get_value(self.I.sel))
            out = BitVector[width](value_store.get_value(self.I.data[int(sel)]))
            value_store.set_value(self.O, out)

        self.simulate = simulate


class Mux(Generator2):
    def __init__(self, height: int, T: Type):
        if issubclass(T, MagmaProtocol):
            T = T._to_magma_()
        T_str = str(T).replace("(", "").replace(")", "")\
                      .replace(",", "_").replace("=", "_")\
                      .replace("[", "").replace("]", "").replace(" ", "")
        self.name = f"Mux{height}x{T_str}"
        # TODO: Type must be hashable so we can cache.
        N = T.flat_length()

        io_dict = {}
        for i in range(height):
            io_dict[f"I{i}"] = In(T)
        if height == 2:
            io_dict["S"] = In(Bit)
        else:
            io_dict["S"] = In(Array[clog2(height), Bit])
        io_dict["O"] = Out(T)

        self.io = IO(**io_dict)
        mux = CoreIRCommonLibMuxN(height, N)()
        data = [Array[N, Bit](getattr(self.io, f"I{i}").flatten())
                for i in range(height)]
        mux.I.data @= Array[height, Array[N, Bit]](data)
        if height == 2:
            mux.I.sel[0] @= self.io.S
        else:
            mux.I.sel @= self.io.S
        self.io.O @= T.unflatten(mux.O.ts)


# NOTE(rsetaluri): We monkeypatch this function on to Array due to the circular
# dependency between Mux and Array. See the discussion on
# https://github.com/phanrahan/magma/pull/658.
def _dynamic_mux_select(this, key):
    return Mux(len(this), this.T)()(*this.ts, key)


Array.dynamic_mux_select = _dynamic_mux_select


def mux(I: list, S, **kwargs):
    """
    How type inference works on I:
        This operator will traverse the list of inputs I and use the first
        magma value to determine the type.  This allows I to contain coerceable
        objects like ints (e.g. `mux([1, x], s)` where `x: UInt[2]`).  Coercion
        is peformed when wiring the arguments to the input of an instance of
        Mux(T) (where T is the type of the first magma value) and will raise an
        error there if one of the values cannot be coerced to T.
        **NOTE** This will fail if the type of the first magma value cannot
        coerce subsequent arguments (even though the type of a later argument
        might be able to coerce the type of the earliest argument).  We plan to
        improve the algorithm to support the more general cases, but until
        then, if you run into this problem, use `Mux(T)` where `T` can coerce
        all arguments.
    """
    if isinstance(S, Type) and S.const():
        S = seq2int(S.bits())
    if isinstance(S, int):
        return I[S]
    # get first magma arg for type introspection
    for arg in I:
        if isinstance(arg, (Type, MagmaProtocol)):
            T = type(arg)
            break
    else:
        raise TypeError("Cannot use m.mux with non-magma types (need at least "
                        "one to infer type)")
    inst = Mux(len(I), T, **kwargs)()
    result = inst(*I, S)
    for i in range(len(I)):
        if getattr(inst, f"I{i}").value() is None:
            arg = I[i]
            raise TypeError(f"mux arg I[{i}] ({arg}: {type(arg)}) does not "
                            f"match inferred input port type {T}")
    return result


# Monkey patch for ite impl without circular dependency
Bit._Mux = Mux



from magma import cache_definition
from magma.backend.coreir_ import CoreIRBackend, CoreIRContextSingleton
from magma.circuit import DefineCircuitKind, Circuit
from magma import cache_definition, Clock, Array, BitIn, BitOut, Product
from coreir.generator import Generator

@cache_definition
def GetCoreIRBackend():
    return CoreIRBackend()

def GetMagmaContext():
    return CoreIRContextSingleton().get_instance()

def ResetCoreIR():
    CoreIRContextSingleton().reset_instance()
    backend = GetCoreIRBackend()
    backend.reset()

_coreirNamedTypeToPortDict = {
    "clk": Clock,
    "coreir.clkIn": Clock
}


def _get_ports_as_list(ports):
    return [item for i in range(len(ports.keys())) for item in
            [list(ports.keys())[i], list(ports.types())[i]]]


def _get_ports(coreir_type, renamed_ports):
    if (coreir_type.kind == "Bit"):
        return BitOut
    elif (coreir_type.kind == "BitIn"):
        return BitIn
    elif (coreir_type.kind == "Array"):
        return Array[len(coreir_type), _get_ports(coreir_type.element_type, renamed_ports)]
    elif (coreir_type.kind == "Record"):
        elements = {}
        for item in coreir_type.items():
            name = item[0]
            # replace  the in port with I as can't reference that
            if name == "in":
                name = "I"
                renamed_ports[name] = "in"
            elements[name] = _get_ports(item[1], renamed_ports)
        return Product.from_fields("anon", elements)
    elif (coreir_type.kind == "Named"):
        # exception to handle clock types, since other named types not handled
        if coreir_type.name in _coreirNamedTypeToPortDict:
            return In(_coreirNamedTypeToPortDict[coreir_type.name])
        else:
            raise NotImplementedError("not all named types supported yet")
    else:
        raise NotImplementedError("Trying to convert unknown coreir type to magma type")


def DefineModuleWrapper(cirb: CoreIRBackend, coreirModule, uniqueName, deps):
    class ModuleWrapper(Circuit):
        name = uniqueName
        renamed_ports = {}
        IO = _get_ports_as_list(_get_ports(coreirModule.type, renamed_ports))
        wrappedModule = coreirModule
        coreir_wrapped_modules_libs_used = set(deps)

        @classmethod
        def definition(cls):
            pass
    return ModuleWrapper

def DefineCircuitFromGeneratorWrapper(cirb: CoreIRBackend, namespace: str, generator: str,
                                      uniqueName: str, dependentNamespaces: list = [],
                                      genargs: dict = {}, runGenerators = True):
    moduleToWrap = cirb.context.import_generator(namespace,generator)(**genargs)

    deps = [namespace] + dependentNamespaces
    if runGenerators:
        cirb.context.run_passes(["rungenerators"], deps)
    print(cirb, cirb.context, cirb.context.context)
    print(moduleToWrap, moduleToWrap.context, moduleToWrap.context.context)
    return DefineModuleWrapper(cirb, moduleToWrap, uniqueName, deps)

def CircuitInstanceFromGeneratorWrapper(cirb: CoreIRBackend, namespace: str, generator: str,
                                        uniqueName: str, dependentNamespaces: list,
                                        genargs: dict = {}, modargs: dict = {}):
    return DefineCircuitFromGeneratorWrapper(cirb, namespace, generator,
                                             uniqueName, dependentNamespaces,
                                             genargs)(**modargs)

def GetCoreIRModule(cirb: CoreIRBackend, circuit: DefineCircuitKind):
    """
    Get the CoreIR module corresponding to the Magma circuit or circuit instance

    :param cirb: The CoreIR backend currently be used.
    :param circuit: The magma circuit to get the coreIR backend for
    :return: A CoreIR module
    """
    if (hasattr(circuit, "wrappedModule")):
        return circuit.wrappedModule
    else:
        # if this is an instance, compile the class, as that is the circuit
        if hasattr(circuit, 'is_instance') and circuit.is_instance:
            circuitNotInstance = circuit.__class__
        else:
            circuitNotInstance = circuit
        moduleOrGenerator = cirb.compile(circuitNotInstance)[circuitNotInstance.name]
        # compile can giv eme back the coreIR module or the coreIR generator. if this is
        # the CoreIR generator, call it with the Magma arguments converted to CoreIR ones.
        if isinstance(moduleOrGenerator, Generator):
            return moduleOrGenerator(**circuitNotInstance.coreir_genargs)
        else:
            return moduleOrGenerator

def DeclareCoreIRGenerator(lib : str, name : str, typegen = None, backend = None):
    if backend is None:
        backend = GetCoreIRBackend()
    if typegen is not None:
        # This is for generators which we don't have access to
        raise NotImplementedError()
    else:
        # Assume the generator is available, create a wrapped circuit
        def Define(**kwargs):
            kwargs_str = "_".join(f"{key}_{value}" for key, value in kwargs.items())
            unique_name = f"{lib}_{name}_{kwargs_str}"
            return DefineCircuitFromGeneratorWrapper(backend, lib, name, unique_name, genargs=kwargs, runGenerators=False)
        return Define

def coreir_typegen(fn):
    def wrapped(*args, **kwargs):
        return None
    return wrapped



from coreir import Context
from ..array import Array
from ..tuple import Tuple
from ..clock import AsyncReset, AsyncResetN, Clock
from ..conversions import convertbit
from ..config import config, EnvConfig
from ..logging import root_logger
from ..t import In, Out, Direction
from .. import singleton
from ..circuit import Circuit
from ..interface import IO
from .coreir_transformer import DefnOrDeclTransformer
from ..passes import DefinitionPass
from .util import keydefaultdict
from ..wire import wire


config._register(
    coreir_backend_log_level=EnvConfig(
        "MAGMA_COREIR_BACKEND_LOG_LEVEL", "WARN"),
)

_logger = root_logger().getChild("coreir_backend")
_logger.setLevel(config.coreir_backend_log_level)


_context_to_modules = {}


# Singleton context meant to be used with coreir/magma code
@singleton
class CoreIRContextSingleton:
    __instance = None

    def get_instance(self):
        return self.__instance

    def reset_instance(self):
        old_instance = self.__instance
        self.__instance = Context()
        if old_instance in _context_to_modules:
            del _context_to_modules[old_instance]
        # Force freeing of C++ memory
        old_instance.delete()

    def __init__(self):
        self.__instance = Context()


CoreIRContextSingleton()


class CoreIRBackend:
    def __init__(self, context=None):
        self._init(context)

    def _init(self, context):
        singleton = CoreIRContextSingleton().get_instance()
        if context is None:
            context = singleton
        if context is not singleton:
            _logger.warning("Creating CoreIRBackend with non-singleton CoreIR "
                            "context.")
        self.modules = _context_to_modules.setdefault(context, {})
        self.context = context
        self.libs = keydefaultdict(self.context.get_lib)
        self.libs_used = set()
        self.constant_cache = {}
        self.sv_bind_files = {}

    def reset(self):
        self._init(context=None)

    def compile(self, defn_or_decl, opts=None):
        _logger.debug(f"Compiling: {defn_or_decl.name}")
        opts = opts if opts is not None else {}
        transformer = DefnOrDeclTransformer(self, opts, defn_or_decl)
        transformer.run()
        self.modules[defn_or_decl.name] = transformer.coreir_module
        return self.modules


class InsertWrapCasts(DefinitionPass):
    def define_wrap(self, wrap_type, in_type, out_type):
        class Wrap(Circuit):
            name = f"coreir_wrap{wrap_type}".replace("(", "").replace(")", "")
            io = IO(**{"in": In(in_type), "out": Out(out_type)})
            coreir_genargs = {"type": wrap_type}
            coreir_name = "wrap"
            coreir_lib = "coreir"

            def simulate(self, value_store, state_store):
                input_val = value_store.get_value(getattr(self, "in"))
                value_store.set_value(self.out, input_val)
        return Wrap

    def wrap_if_named_type(self, port, definition):
        if isinstance(port, Tuple):
            wrapped = False
            for t in port:
                wrapped |= self.wrap_if_named_type(t, definition)
            return wrapped
        if isinstance(port, Array):
            wrapped = self.wrap_if_named_type(port[0], definition)
            if not wrapped:
                return False
            # TODO: Magma doesn't support length zero array, so slicing a
            # length 1 array off the end doesn't work as expected in normal
            # Python, so we explicilty slice port.ts
            for t in port.ts[1:]:
                self.wrap_if_named_type(t, definition)
            return True
        if not port.is_input():
            return False
        value = port.value()
        if not (isinstance(port, (AsyncReset, AsyncResetN, Clock)) or
                isinstance(value, (AsyncReset, AsyncResetN, Clock))):
            return False
        undirected_t = type(port).qualify(Direction.Undirected)
        if value is None or issubclass(type(value), undirected_t):
            return False
        if isinstance(port, (AsyncReset, AsyncResetN, Clock)):
            T = Out(type(port))
        else:
            T = In(type(value))

        with definition.open():
            port.unwire(value)
            inst = self.define_wrap(T, type(port), type(value))()
            wire(convertbit(value, type(port)), getattr(inst, "in"))
            wire(inst.out, convertbit(port, type(value)))
        return True

    def __call__(self, definition):
        # copy, because wrapping might add instances
        instances = definition.instances[:]
        for instance in definition.instances:
            if type(instance).coreir_name == "wrap" or \
                    type(instance).coreir_name == "unwrap":
                continue
            for port in instance.interface.ports.values():
                self.wrap_if_named_type(port, definition)
        for port in definition.interface.ports.values():
            self.wrap_if_named_type(port, definition)


def compile(main, file_name=None, context=None):
    InsertWrapCasts(main).run()
    backend = CoreIRBackend(context)
    backend.compile(main)
    if file_name is not None:
        return backend.modules[main.coreir_name].save_to_file(file_name)
    else:
        return backend.modules[main.coreir_name]



import ast
import enum
import textwrap
import inspect
from functools import wraps
import functools
import operator
from collections import namedtuple
import os

import six
from . import cache_definition
from .common import deprecated, setattrs, Stack, OrderedIdentitySet
from .interface import *
from .wire import *
from .config import get_debug_mode
from .debug import get_callee_frame_info, debug_info
from .logging import root_logger, stage_logger, unstage_logger
from .is_definition import isdefinition
from .placer import Placer, StagedPlacer
from magma.syntax.combinational import combinational
from magma.syntax.sequential import sequential
try:
    import kratos
    from magma.syntax.verilog import combinational_to_verilog, \
        sequential_to_verilog
except ImportError:
    pass
from .view import PortView
from magma.protocol_type import MagmaProtocol
from magma.wire_container import WiringLog

__all__ = ['AnonymousCircuitType']
__all__ += ['AnonymousCircuit']

__all__ += ['CircuitType']
__all__ += ['Circuit']
__all__ += ['DeclareCircuit']
__all__ += ['DefineCircuit', 'EndDefine', 'EndCircuit']
__all__ += ['DefineCircuitKind']

__all__ += ['CopyInstance']
__all__ += ['circuit_type_method']
__all__ += ['circuit_generator']
__all__ += ['CircuitBuilder', 'builder_method']

circuit_type_method = namedtuple('circuit_type_method', ['name', 'definition'])

_logger = root_logger()

_VERILOG_FILE_OPEN = """
integer \_file_{filename} ;
initial \_file_{filename} = $fopen(\"{filename}\", \"{mode}\");
"""  # noqa

_VERILOG_FILE_CLOSE = """
final $fclose(\_file_{filename} );
"""  # noqa

_DEFAULT_VERILOG_LOG_STR = f"""
`ifndef MAGMA_LOG_LEVEL
    `define MAGMA_LOG_LEVEL 1
`endif"""


class _SyntaxStyle(enum.Enum):
    NONE = enum.auto()
    OLD = enum.auto()
    NEW = enum.auto()


class DefinitionContext:
    def __init__(self, placer):
        stage_logger()
        self.placer = placer
        self._inline_verilog = []
        self._displays = []
        self._insert_default_log_level = False
        self._files = []
        self._builders = []

    def add_builder(self, builder):
        self._builders.append(builder)

    def add_file(self, file):
        self._files.append(file)

    def _finalize_file_opens(self):
        for file in self._files:
            self.add_inline_verilog(
                _VERILOG_FILE_OPEN.format(filename=file.filename,
                                          mode=file.mode),
                {}, {})

    def _finalize_file_close(self):
        for file in self._files:
            self.add_inline_verilog(
                _VERILOG_FILE_CLOSE.format(filename=file.filename),
                {}, {})

    def add_inline_verilog(self, format_str, format_args, symbol_table):
        self._inline_verilog.append((format_str, format_args, symbol_table))

    def insert_default_log_level(self):
        self._insert_default_log_level = True

    def add_display(self, display):
        self._displays.append(display)

    def _finalize_displays(self):
        for display in self._displays:
            self.add_inline_verilog(*display.get_inline_verilog())

    def finalize(self, defn):
        self.placer = self.placer.finalize(defn)
        for builder in self._builders:
            inst = builder.finalize()
            self.placer.place(inst)
        if self._insert_default_log_level:
            self.add_inline_verilog(_DEFAULT_VERILOG_LOG_STR, {}, {})
        self._finalize_file_opens()  # so displays can refer to open files
        self._finalize_displays()
        self._finalize_file_close()  # close after displays
        unstage_logger()


_definition_context_stack = Stack()


class _DefinitionContextManager:
    def __init__(self, context):
        self._context = context

    def __enter__(self):
        _definition_context_stack.push(self._context)

    def __exit__(self, typ, value, traceback):
        _definition_context_stack.pop()


def _has_definition(cls, port=None):
    if cls.instances:
        return True
    if cls._context_._inline_verilog:
        return True
    if port is None:
        interface = getattr(cls, "interface", None)
        if not interface:
            return False
        return any(_has_definition(cls, p) for p in interface.ports.values())
    flat = port.flatten()
    return any(not f.is_output() and f.value() is not None for f in flat)


def _get_interface_type(cls):
    if hasattr(cls, "IO") and hasattr(cls, "io"):
        _logger.warning("'IO' and 'io' should not both be specified, ignoring "
                        "'io'", debug_info=cls.debug_info)
    # TODO(setaluri): Simplify this logic and remove InterfaceKind check
    # (this is a artifact of the circuit_generator decorator).
    if hasattr(cls, 'IO'):
        cls._syntax_style_ = _SyntaxStyle.OLD
        _logger.warning("'IO = [...]' syntax is deprecated, use "
                        "'io = IO(...)' syntax instead",
                        debug_info=cls.debug_info)
        if isinstance(cls.IO, InterfaceKind):
            return None
        return make_interface(*cls.IO)
    if hasattr(cls, "io"):
        cls._syntax_style_ = _SyntaxStyle.NEW
        return cls.io.make_interface()
    return None


def _add_intermediate_value(value):
    """
    Add an intermediate value `value` to `values`, handling members of
    recursive types.  Used by `_get_intermediate_values` as part of
    `CircuitKind.__repr__`.
    """
    root = value.name.root()
    if root is None or root.const():
        return OrderedIdentitySet()
    return OrderedIdentitySet((root,))


def _get_intermediate_values(value):
    """
    Retrieve the intermediate values connected to `value` and put them into
    `values`

    Used in the implementation of `CircuitKind.__repr__` to emit the temporary
    values in a circuit.
    """
    if value.is_output():
        return OrderedIdentitySet()
    if value.is_mixed():
        # Mixed
        return functools.reduce(operator.or_,
                                (_get_intermediate_values(v) for v in value),
                                OrderedIdentitySet())
    driver = value.value()
    if driver is None:
        return OrderedIdentitySet()
    flat = value.flatten()
    if len(flat) > 1 and driver.name.anon():
        return functools.reduce(operator.or_,
                                (_get_intermediate_values(f) for f in flat),
                                OrderedIdentitySet())
    values = OrderedIdentitySet()
    while driver is not None:
        values |= _add_intermediate_value(driver)
        if not driver.is_input():
            break
        value = driver
        driver = driver.value()
    return values


class CircuitKind(type):
    def __prepare__(name, bases, **kwargs):
        context = DefinitionContext(StagedPlacer(name))
        _definition_context_stack.push(context)
        return type.__prepare__(name, bases, **kwargs)

    """Metaclass for creating circuits."""
    def __new__(metacls, name, bases, dct):
        # Override class name if supplied (and save class name).
        cls_name = dct.get("_cls_name_", name)
        name = dct.setdefault('name', name)

        dct.setdefault('renamed_ports', {})
        dct.setdefault('primitive', False)
        dct.setdefault('coreir_lib', 'global')
        dct.setdefault("inline_verilog_strs", [])
        dct["inline_verilog_generated"] = False
        dct["bind_modules"] = {}
        dct["bind_modules_bound"] = False

        # If in debug_mode is active and debug_info is not supplied, attach
        # callee stack info.
        dct.setdefault("debug_info", None)
        if get_debug_mode() and not dct["debug_info"]:
            callee_frame = inspect.getframeinfo(
                inspect.currentframe().f_back.f_back)
            module = inspect.getmodule(inspect.stack()[2][0])
            dct["debug_info"] = debug_info(callee_frame.filename,
                                           callee_frame.lineno, module)

        cls = type.__new__(metacls, name, bases, dct)

        for method in dct.get('circuit_type_methods', []):
            setattr(cls, method.name, method.definition)

        # Create interface for this circuit class.
        cls._syntax_style_ = _SyntaxStyle.NONE
        IO = _get_interface_type(cls)
        if IO is not None:
            cls.IO = IO
            cls.interface = cls.IO(defn=cls, renamed_ports=dct["renamed_ports"])
            setattrs(cls, cls.interface.ports, lambda k, v: isinstance(k, str))
        try:
            context = _definition_context_stack.pop()
            assert context.placer.name == cls_name
            # Override staged context with '_context_' from namespace if
            # available.
            cls._context_ = dct.get("_context_", context)
            cls._context_.finalize(cls)
        except IndexError:  # no staged placer
            cls._context_ = DefinitionContext(Placer(cls))

        return cls

    def __call__(cls, *largs, **kwargs):
        if get_debug_mode():
            debug_info = get_callee_frame_info()
            kwargs["debug_info"] = debug_info
        self = super(CircuitKind, cls).__call__(*largs, **kwargs)
        if hasattr(cls, 'IO'):
            interface = cls.IO(inst=self, renamed_ports=cls.renamed_ports)
            self.setinterface(interface)
        return self

    def __str__(cls):
        interface = ""
        if hasattr(cls, "interface"):
            interface = ", ".join(f"{name}: {_type}"
                                  for name, _type in cls.interface.items())
            interface = f"({interface})"
        return f"{cls.__name__}{interface}"

    def __repr__(cls):
        if not hasattr(cls, 'IO'):
            return super().__repr__()

        name = cls.__name__
        args = str(cls.IO)

        if not isdefinition(cls):
            return f"{name} = DeclareCircuit(\"{name}\", {args})"
        s = f"{name} = DefineCircuit(\"{name}\", {args})\n"
        sorted_instances = sorted(cls.instances, key=lambda x: x.name)

        values = []
        for instance in sorted_instances:
            values.extend(list(instance.interface.ports.values()))
        values.extend(cls.interface.ports.values())
        intermediate_values = (_get_intermediate_values(value)
                               for value in values)
        intermediate_values = functools.reduce(operator.or_,
                                               intermediate_values,
                                               OrderedIdentitySet())
        if intermediate_values:
            intermediate_values = (f"{value.name} = {repr(value)}"
                                   for value in intermediate_values)
            s += "\n".join(intermediate_values) + "\n"
        # Emit instances.
        for instance in sorted_instances:
            s += repr(instance) + '\n'
        # Emit wires from instances.
        for instance in sorted_instances:
            s += repr(instance.interface)
        s += repr(cls.interface)
        s += "EndCircuit()"
        return s

    def rename(cls, new_name):
        if cls.verilogFile:
            raise Exception("Can not rename a verilog wrapped file")

        old_name = cls.name
        cls.name = new_name
        cls.coreir_name = new_name
        cls.verilog_name = new_name
        cls.__name__ = new_name

        # NOTE(rsetaluri): This is a very hacky way to try to rename wrapped
        # verilog. We simply replace the first instance of "module <old_name>"
        # with "module <new_name>". This ignores the possibility of "module
        # <new_name>" existing anywhere else, most likely in comments etc. The
        # more robust way to do this would to modify the AST directly and
        # generate the new verilog code.
        #
        # Instead, at the top of this function, we raise an exception for
        # verilog wrapped files. That is, for now it is considered an error to
        # try to rename a verilog wrapped circuit.
        if cls.verilogFile:
            find_str = f"module {old_name}"
            replace_str = f"module {new_name}"
            assert cls.verilogFile.find(find_str) != -1
            cls.verilogFile = cls.verilogFile.replace(find_str, replace_str, 1)

    def find(cls, defn):
        name = cls.__name__
        if not isdefinition(cls):
            return defn
        for i in cls.instances:
            type(i).find(defn)
        if name not in defn:
            defn[name] = cls
        return defn

    @deprecated(msg="cls.inline_verilog is deprecated, use m.inline_verilog"
                "instead")
    def inline_verilog(cls, inline_str, **kwargs):
        cls._context_.add_inline_verilog(inline_str, kwargs, None)


@six.add_metaclass(CircuitKind)
class AnonymousCircuitType(object):
    """Abstract base class for circuits"""
    def __init__(self, *largs, **kwargs):
        self.kwargs = dict(**kwargs)

        if "debug_info" in self.kwargs:
            # Not an instance parameter, internal debug argument.
            del self.kwargs["debug_info"]

        if hasattr(self, 'default_kwargs'):
            for key, value in self.default_kwargs.items():
                self.kwargs.setdefault(key, value)

        self.name = kwargs['name'] if 'name' in kwargs else ""
        self.loc = kwargs['loc'] if 'loc' in kwargs else None
        if self.loc and len(self.loc) == 2:
            self.loc = (self.loc[0], self.loc[1], 0)
        self.interface = None
        self.defn = None
        self.used = False
        self.is_instance = True
        self.debug_info = kwargs.get("debug_info", None)

    def set_debug_info(self, debug_info):
        self.debug_info = debug_info

    def __str__(self):
        if self.name:
            return f"{self.name}<{type(self)}>"
        name = f"AnonymousCircuitInst{id(self)}"
        interface = ", ".join(f"{name}: {type(value)}"
                              for name, value in self.interface.ports.items())
        return f"{name}<{interface}>"

    def __repr__(self):
        args = []
        for k, v in self.interface.ports.items():
            args.append(f"\"{k}\"")
            args.append(repr(v))
        args = ", ".join(args)
        typ = type(self).__name__
        if self.name:
            return f"{self.name} = {typ}({args})"
        return f"{typ}({args})"

    def __getitem__(self, key):
        return self.interface[key]

    def wireoutputs(self, outputs, debug_info):
        """Wire a list of outputs to the circuit's inputs"""
        inputs = self.interface.inputs()
        ni = len(inputs)
        no = len(outputs)
        if ni != no:
            msg = (
                f"Number of inputs is not equal to the number of outputs, "
                f"expected {ni} inputs, got {no}. Only {min(ni,no)} will be "
                f"wired.")
            _logger.warning(msg, debug_info=debug_info)
        for i in range(min(ni, no)):
            wire(outputs[i], inputs[i], debug_info)

    def wire(self, output, debug_info):
        """Wire a single output to the circuit's inputs"""
        if hasattr(output, 'interface'):
            # Wire the circuit's outputs to this circuit's inputs.
            self.wireoutputs(output.interface.outputs(), debug_info)
            return
        if isinstance(output, MagmaProtocol):
            output = output._get_magma_value_()
        # Wire the output to this circuit's input (should only have 1 input).
        inputs = []
        for inp in self.interface.inputs():
            if isinstance(inp, MagmaProtocol):
                inp = inp._get_magma_value_()
            inputs.append(inp)
        ni = len(inputs)
        if ni == 0:
            msg = ("Wiring an output to a circuit with no input arguments, "
                   "skipping")
            _logger.warning(msg, debug_info=debug_info)
            return
        if ni != 1:
            _logger.warning(
                WiringLog("Wiring an output to a circuit with more than one "
                          "input argument, using the first input {}",
                          inputs[0]),
                debug_info=debug_info
            )
        inputs[0].wire(output, debug_info)

    @property
    def debug_name(self):
        defn_str = ""
        if hasattr(self, 'defn') and self.defn is not None:
            defn_str = str(self.defn.name)
        return f"{defn_str}.{self.name}"

    def __call__(input, *outputs, **kw):
        debug_info = None
        if get_debug_mode():
            debug_info = get_callee_frame_info()

        no = len(outputs)
        if len(outputs) == 1:
            input.wire(outputs[0], debug_info)
        elif len(outputs) >= 1:  # In case there are only kw
            input.wireoutputs(outputs, debug_info)

        # Wire up extra arguments, name to name.
        # TODO(rsetaluri): This code should be changed to use clock types.
        for key, value in kw.items():
            if key == 'enable':
                key = 'CE'
            elif key == 'reset':
                key = 'RESET'
            elif key == 'set':
                key = 'SET'  # NYI
            elif key == 'ce':
                key = 'CE'  # deprecated
            if hasattr(input, key):
                i = getattr(input, key)
                wire(value, getattr(input, key), debug_info)
            else:
                _logger.warning(
                    WiringLog(f"Instance {{}} does not have input {key}",
                              input),
                    debug_info=debug_info
                )

        o = input.interface.outputs()
        return o[0] if len(o) == 1 else tuple(o)

    def setinterface(self, interface):
        setattrs(self, interface.ports, lambda k, v: isinstance(k, str))
        self.interface = interface
        return self

    def on(self):
        self.used = True
        return self

    def off(self):
        self.used = False
        return self

    def rename(self, name):
        self.name = name
        return self

    def isclocked(self):
        return self.interface.isclocked()

    def clockargs(self):
        return self.interface.clockargs()

    def inputargs(self):
        return self.interface.inputargs()

    def outputargs(self):
        return self.interface.outputargs()

    @classmethod
    def open(cls):
        return _DefinitionContextManager(cls._context_)


def AnonymousCircuit(*decl):
    """
    AnonymousCircuits are like macros - the circuit instances are not placed
    """
    if len(decl) == 1:
        decl = decl[0]
    return AnonymousCircuitType().setinterface(AnonymousInterface(decl))


class CircuitType(AnonymousCircuitType):
    """Placed circuit - instances placed in a definition"""
    def __init__(self, *largs, **kwargs):
        super(CircuitType, self).__init__(*largs, **kwargs)
        try:
            context = _definition_context_stack.peek()
            context.placer.place(self)
        except IndexError:  # instances must happen inside a definition context
            raise Exception("Can not instance a circuit outside a definition")

    def __repr__(self):
        args = []
        for k, v in self.kwargs.items():
            if isinstance(v, tuple):
                # {   # Format identifier
                # 0:  # first parameter
                # #   # use "0x" prefix
                # 0   # fill with zeroes
                # {1} # to n characters (including 0x), per the second parameter
                # x   # hexadecimal number, using lowercase letters for a-f
                # }   # End of format identifier
                if len(v) == 2:
                    v = "{0:#0{1}x}".format(v[0], v[1] // 4)
            else:
                v = f"\"{v}\""
            args.append(f"{k}={v}")
        args = ", ".join(args)
        typ = type(self).__name__
        if self.name:
            return f"{self.name} = {typ}({args})"
        return f"{typ}({args})"


@deprecated(
    msg="DeclareCircuit factory method is deprecated, subclass Circuit instead")
def DeclareCircuit(name, *decl, **args):
    """DeclareCircuit Factory"""
    debug_info = None
    if get_debug_mode():
        debug_info = get_callee_frame_info()
    metacls = CircuitKind
    bases = (CircuitType,)
    dct = metacls.__prepare__(name, bases)
    dct.update(dict(IO=decl,
                    debug_info=debug_info,
                    is_definition=False,
                    primitive=args.get('primitive', True),
                    stateful=args.get('stateful', False),
                    simulate=args.get('simulate'),
                    firrtl_op=args.get('firrtl_op'),
                    circuit_type_methods=args.get('circuit_type_methods', []),
                    coreir_lib=args.get('coreir_lib', "global"),
                    coreir_name=args.get('coreir_name', name),
                    coreir_genargs=args.get('coreir_genargs', None),
                    coreir_configargs=args.get('coreir_configargs', {}),
                    verilog_name=args.get('verilog_name', name),
                    default_kwargs=args.get('default_kwargs', {}),
                    renamed_ports=args.get('renamed_ports', {})))
    return metacls(name, bases, dct)


class DefineCircuitKind(CircuitKind):
    def __new__(metacls, name, bases, dct):
        dct["_cls_name_"] = name  # save original name for debugging purposes
        if 'name' not in dct:
            dct['name'] = name
            # Check if we are a subclass of something other than Circuit; if so,
            # inherit the name of the first parent.
            for base in bases:
                if base is Circuit:
                    continue
                if not issubclass(base, AnonymousCircuitType):
                    raise Exception(f"Must subclass from AnonymousCircuitType "
                                    f"or a subclass of AnonymousCircuitType "
                                    f"({base})")
                dct['name'] = base.name
                break
        name = dct['name']
        dct["renamed_ports"] = dct.get("renamed_ports", {})

        self = CircuitKind.__new__(metacls, name, bases, dct)

        self.verilog = None
        self.verilogFile = None
        self.verilogLib = None

        self.verilog_name = dct.get('verilog_name', name)
        self.coreir_name = dct.get('coreir_name', name)
        self.coreir_genargs = dct.get('coreir_genargs', None)
        self.coreir_configargs = dct.get('coreir_configargs', {})
        self.default_kwargs = dct.get('default_kwargs', {})
        self.firrtl = None

        has_definition = _has_definition(self)
        self._is_definition = dct.get('is_definition', has_definition)
        self.is_instance = False

        run_unconnected_check = False
        if hasattr(self, "definition"):
            if self._syntax_style_ is _SyntaxStyle.OLD:
                _logger.warning("'definition' class method syntax is "
                                "deprecated, use inline definition syntax "
                                "instead", debug_info=self.debug_info)
                with _DefinitionContextManager(self._context_):
                    self.definition()
                self._is_definition = True
                run_unconnected_check = True
            elif self._syntax_style_ is _SyntaxStyle.NEW:
                _logger.warning("Supplying method 'definition' with new inline "
                                "definition syntax is not supported, ignoring "
                                "'definition'")
        run_unconnected_check = run_unconnected_check or (
            has_definition and self._syntax_style_ is _SyntaxStyle.NEW)
        run_unconnected_check = run_unconnected_check and not \
            dct.get("_ignore_undriven_", False)
        if run_unconnected_check:
            self.check_unconnected()

        return self

    def check_unconnected(self):
        for port in self.interface.inputs():
            if not port.driven():
                msg = f"Output port {self.name}.{port.name} not driven"
                _logger.error(msg, debug_info=self.debug_info)

        for inst in self.instances:
            for port in inst.interface.inputs():
                if not port.driven():
                    msg = f"Input port {inst.name}.{port.name} not driven"
                    _logger.error(msg, debug_info=inst.debug_info)

    @property
    def is_definition(self):
        return self._is_definition or self.verilog or self.verilogFile

    @property
    def instances(self):
        return self._context_.placer.instances()

    def place(cls, inst):
        """Place a circuit instance in this definition"""
        cls._context_.placer.place(inst)

    def bind(cls, monitor, *args):
        cls.bind_modules[monitor] = args


@six.add_metaclass(DefineCircuitKind)
class Circuit(CircuitType):
    pass


@deprecated(
    msg="DefineCircuit factory method is deprecated, subclass Circuit instead")
def DefineCircuit(name, *decl, **args):
    """DefineCircuit Factory"""
    debug_info = None
    if get_debug_mode():
        debug_info = get_callee_frame_info()
    metacls = DefineCircuitKind
    bases = (Circuit,)
    dct = metacls.__prepare__(name, bases)
    dct.update(dict(IO=decl,
                    is_definition=True,
                    primitive=args.get('primitive', False),
                    stateful=args.get('stateful', False),
                    simulate=args.get('simulate'),
                    debug_info=debug_info,
                    verilog_name=args.get('verilog_name', name),
                    coreir_name=args.get('coreir_name', name),
                    coreir_lib=args.get('coreir_lib', "global"),
                    coreir_genargs=args.get('coreir_genargs', None),
                    coreir_configargs=args.get('coreir_configargs', None),
                    default_kwargs=args.get('default_kwargs', {}),
                    renamed_ports=args.get('renamed_ports', {}),
                    kratos=args.get("kratos", None)))
    defn = metacls(name, bases, dct)
    _definition_context_stack.push(defn._context_)
    return defn


def EndDefine():
    try:
        context = _definition_context_stack.pop()
    except IndexError:
        raise Exception("EndDefine not matched to DefineCircuit")
    placer = context.placer
    placer._defn.check_unconnected()
    debug_info = get_callee_frame_info()
    placer._defn.end_circuit_filename = debug_info[0]
    placer._defn.end_circuit_lineno = debug_info[1]


EndCircuit = EndDefine


def CopyInstance(instance):
    circuit = type(instance)
    new_instance = circuit()
    new_instance.kwargs = instance.kwargs
    return new_instance


_GeneratorArguments = namedtuple('GeneratorArguments', ['args', 'kwargs'])


@deprecated(msg="circuit_generator decorator is deprecated, subclass Generator "
            "instead")
def circuit_generator(func):
    @cache_definition
    @wraps(func)
    def wrapped(*args, **kwargs):
        result = func(*args, **kwargs)
        # Store arguments to generate the circuit.
        result._generator_arguments = _GeneratorArguments(args, kwargs)
        return result
    return wrapped


coreir_port_mapping = {
    "I": "in",
    "I0": "in0",
    "I1": "in1",
    "O": "out",
    "S": "sel",
    "CLK": "clk",
}


def DeclareCoreirCircuit(*args, **kwargs):
    return DeclareCircuit(*args, **kwargs,
                          renamed_ports=coreir_port_mapping)


class _CircuitBuilderMeta(type):
    pass


def builder_method(func):

    @wraps(func)
    def _wrapped(this, *args, **kwargs):
        with _DefinitionContextManager(this._context):
            result = func(this, *args, **kwargs)
        return result

    return _wrapped


class CircuitBuilder(metaclass=_CircuitBuilderMeta):
    _RESERVED_NAMESPACE_KEYS = {"io", "_context_", "name"}

    def __init__(self, name):
        try:
            context = _definition_context_stack.peek()
            context.add_builder(self)
        except IndexError:
            raise Exception("Can not instance a circuit builder outside a "
                            "definition")
        self._name = name
        self._io = SingletonInstanceIO()
        self._finalized = False
        self._dct = {}
        self._context = DefinitionContext(StagedPlacer(self._name))

    def _port(self, name):
        return self._io.ports[name]

    def _add_port(self, name, typ):
        self._io.add(name, typ)
        setattr(self, name, self._io.inst_ports[name])

    def _set_namespace_key(self, key, value):
        if key in CircuitBuilder._RESERVED_NAMESPACE_KEYS:
            raise Exception(f"Can not set reserved namespace key '{key}'")
        self._dct[key] = value

    def _finalize(self):
        pass

    def finalize(self):
        if self._finalized:
            raise Exception("Can only call finalize on a CircuitBuilder once")
        self._finalize()
        bases = (AnonymousCircuitType,)
        dct = {"io": self._io, "_context_": self._context, "name": self._name}
        dct.update(self._dct)
        DefineCircuitKind.__prepare__(self._name, bases)
        t = DefineCircuitKind(self._name, bases, dct)
        self._finalized = True
        return t(name=self._name)



import ast
import magma as magma_module
from typing import Optional, MutableMapping

from ast_tools.stack import SymbolTable
from ast_tools.passes import (PASS_ARGS_T, Pass, apply_ast_passes, ssa,
                              if_to_phi, bool_to_bit)

from ..circuit import Circuit, IO
from ..clock import AbstractReset
from ..clock_io import ClockIO, gen_clock_io
from ..t import Type, Direction
from ..protocol_type import MagmaProtocol, MagmaProtocolMeta
from ..primitives.register import Register

from .util import build_io_args, build_call_args, wire_call_result


class _SequentialListWrapper(list):
    def __getitem__(self, i):
        result = super().__getitem__(i)
        if isinstance(type(result), Register):
            # Patch in prev method as syntax sugar for previous value
            result.prev = lambda: result.O
        if isinstance(result, list):
            return _SequentialListWrapper(result)
        return result

    def __setitem__(self, i, value):
        item = self[i]
        if not isinstance(item, Circuit):
            raise TypeError("Excepted setitem target to be a Circuit")

        if not isinstance(value, (Circuit, int, Type)):
            raise TypeError("Excepted setitem value to be a Circuit, Type, or "
                            f"int, not {type(value)}")

        item(value)


class _SequentialRegisterWrapperMeta(MagmaProtocolMeta):
    def _to_magma_(cls):
        return cls.T

    def _qualify_magma_(cls, direction: Direction):
        return cls[cls.T.qualify(direction)]

    def _flip_magma_(cls):
        return cls[cls.T.flip()]

    def _from_magma_value_(cls, val: Type):
        return cls(val)

    def __getitem__(cls, T):
        return type(cls)(f"_SequentialRegisterWrapper{T}", (cls, ), {"T": T})


class _SequentialRegisterWrapper(MagmaProtocol,
                                 metaclass=_SequentialRegisterWrapperMeta):
    """
    Wraps sequential attributes to implement register getattr syntax and
    general circuit call syntax, e.g.

        def __init__(self):
            self.x = Register(4)

        def __call__(self, ...) -> ...:
            if S:
                return self.x(I)  # call syntax
            return self.x + 1     # attribute syntax
    """

    def __init__(self, circuit: Circuit):
        self.circuit = circuit

    def _get_magma_value_(self):
        # Return output
        return self.circuit()

    # TODO: Clever way to pass through operators to the underlying value?
    def __add__(self, other):
        return self._get_magma_value_() + other

    def __radd__(self, other):
        return other + self._get_magma_value_()

    def __sub__(self, other):
        return self._get_magma_value_() - other

    def __rsub__(self, other):
        return other - self._get_magma_value_()

    def __mul__(self, other):
        return self._get_magma_value_() * other

    def __rmul__(self, other):
        return other * self._get_magma_value_()

    def __truediv__(self, other):
        return self._get_magma_value_() / other

    def __rtruediv__(self, other):
        return other / self._get_magma_value_()

    def __floordiv__(self, other):
        return self._get_magma_value_() // other

    def __rfloordiv__(self, other):
        return other // self._get_magma_value_()

    def __mod__(self, other):
        return self._get_magma_value_() % other

    def __rmod__(self, other):
        return other % self._get_magma_value_()

    def __or__(self, other):
        return self._get_magma_value_() | other

    def __ror__(self, other):
        return other | self._get_magma_value_()

    def __xor__(self, other):
        return self._get_magma_value_() ^ other

    def __rxor__(self, other):
        return other ^ self._get_magma_value_()

    def __and__(self, other):
        return self._get_magma_value_() & other

    def __rand__(self, other):
        return other & self._get_magma_value_()

    def __eq__(self, other):
        return self._get_magma_value_() == other

    def __ne__(self, other):
        return self._get_magma_value_() != other

    def __ge__(self, other):
        return self._get_magma_value_() >= other

    def __gt__(self, other):
        return self._get_magma_value_() > other

    def __le__(self, other):
        return self._get_magma_value_() <= other

    def __lt__(self, other):
        return self._get_magma_value_() < other

    def __lshift__(self, other):
        return self._get_magma_value_() << other

    def __rlshift__(self, other):
        return other << self._get_magma_value_()

    def __rshift__(self, other):
        return self._get_magma_value_() >> other

    def __rrshift__(self, other):
        return other >> self._get_magma_value_()

    def __neg__(self, other):
        return -self._get_magma_value_()

    def __invert__(self, other):
        return ~self._get_magma_value_()

    def ite(self, s, t, f):
        return self._get_magma_value_().ite(s, t, f)

    def __call__(self, *args, **kwargs):
        return self.circuit(*args, **kwargs)

    def __getitem__(self, i):
        key = i._get_magma_value_() if isinstance(i, MagmaProtocol) else i
        return self._get_magma_value_()[key]

    def __len__(self):
        return len(self._get_magma_value_())


def sequential_getattribute(self, key):
    """
    Wrap lists so we can use the setattr interface with lists of registers
    """
    result = object.__getattribute__(self, key)
    if isinstance(result, list):
        return _SequentialListWrapper(result)
    if isinstance(type(result), Register):
        return _SequentialRegisterWrapper[type(result())](result)
    return result


def sequential_setattr(self, key, value):
    # TODO: for now we assume this is a register, ideally we'd type check this,
    # but we need to have a notion of a register primitive (e.g. right now the
    # mantle reg wraps the coreir reg primitive, so technically the register is
    # an arbitrary user-defined circuit)

    target = object.__getattribute__(self, key)
    # We can at least match the value is a circuit with outputs that match the
    # input of the attribute circuit
    if not isinstance(target, Circuit):
        raise TypeError("Excepted setattr key to be a Circuit")

    if isinstance(value, MagmaProtocol):
        value = value._get_magma_value_()
    if not isinstance(value, (Circuit, int, Type)):
        raise TypeError("Excepted setattr value to be a Circuit, Type, or int",
                        f"not {type(value)}")

    # Simply use function call syntax for now (should automatically retrieve
    # the output of value)
    target(value)


def _process_phi_arg(arg):
    if isinstance(arg, Circuit):
        # Call with no args to retrieve output(s)
        arg = arg()
        if isinstance(arg, list):
            # TODO: We could support tuples/products
            raise TypeError("Cannot use circuit with multiple outputs as "
                            "argument to phi")
    return arg


def _seq_phi(s, t, f):
    if isinstance(t, tuple):
        assert isinstance(f, tuple)
        return tuple(_seq_phi(s, _t, _f) for _t, _f in zip(t, f))
    # Unpack registers/circuits used in getattr syntax
    t, f = map(_process_phi_arg, (t, f))
    return s.ite(t, f)


class _RegisterUpdater(ast.NodeTransformer):
    """Update m.Register params implicitly"""
    def __init__(self, env, reset_type, has_enable, reset_priority):
        self.env = env

        for k, v in env.items():
            if v is magma_module:
                magma_id = k
                break
        if not magma_id:
            raise Exception("Cannot find magma module")

        self.reset_type = ast.keyword(
            arg="reset_type",
            value=ast.parse(f"{magma_id}.{reset_type}").body[0].value,
        )
        self.has_enable = ast.keyword(
            arg="has_enable",
            value=ast.Constant(has_enable),
        )
        self.reset_priority = ast.keyword(
            arg="reset_priority",
            value=ast.Constant(reset_priority),
        )

    def visit_Call(self, node: ast.Call):
        # find m.Register
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.attr == "Register"
            and eval(node.func.value.id, {}, self.env) is magma_module
        ):
            # set params when not exists
            keywords = []
            args = set()
            for keyword in node.keywords:
                keywords.append(keyword)
                args.add(keyword.arg)

            if "reset_type" not in args:
                keywords.append(self.reset_type)
            if "has_enable" not in args:
                keywords.append(self.has_enable)
            if "reset_priority" not in args:
                keywords.append(self.reset_priority)

            return ast.Call(func=node.func, args=node.args, keywords=keywords)

        return self.generic_visit(node)


class _UpdateRegister(Pass):
    """Update m.Register params implicitly"""
    def __init__(self, reset_type, has_enable, reset_priority):
        self.reset_type = reset_type
        self.has_enable = has_enable
        self.reset_priority = reset_priority

    def rewrite(
        self, tree: ast.AST, env: SymbolTable, metadata: MutableMapping
    ) -> PASS_ARGS_T:
        updater = _RegisterUpdater(env, self.reset_type,
                                   self.has_enable, self.reset_priority)
        return updater.visit(tree), env, metadata


class _PrevReplacer(ast.NodeTransformer):
    """Replace self.attr.prev() with __magma_sequential2_attr"""
    def __init__(self):
        self.attrs = {}

    def visit_Call(self, node: ast.Call):
        # find self.attr.prev()
        if (isinstance(node.func, ast.Attribute) and
                isinstance(node.func.value, ast.Attribute) and
                isinstance(node.func.value.value, ast.Name) and
                node.func.value.value.id == "self" and
                node.func.attr == "prev"):
            # replace self.attr.prev() with __magma_sequential2_attr
            attr = node.func.value
            var = f"__magma_sequential2_{attr.attr}"
            self.attrs[var] = attr
            return ast.Name(id=var, ctx=ast.Load())
        return self.generic_visit(node)


class _ReplacePrev(Pass):
    """
    Replace all self.attr.prev() usage

    def __call__(self):
        return self.attr.prev()

    def __call__(self):
        __magma_sequential2_attr = self.attr
        return __magma_sequential2_attr
    """

    def rewrite(self,
                tree: ast.AST,
                env: SymbolTable,
                metadata: MutableMapping) -> PASS_ARGS_T:
        prev_replacer = _PrevReplacer()
        tree = prev_replacer.visit(tree)

        assigns = []
        for var, attr in prev_replacer.attrs.items():
            assigns.append(ast.Assign(targets=[
                ast.Name(id=var, ctx=ast.Store())
            ], value=attr))
        tree.body = assigns + tree.body

        return tree, env, metadata


def sequential2(pre_passes=[], post_passes=[],
                debug: bool = False,
                env: Optional[SymbolTable] = None,
                path: Optional[str] = None,
                file_name: Optional[str] = None,
                annotations: Optional[dict] = None,
                reset_type: AbstractReset = None,
                has_enable: bool = False,
                reset_priority: bool = True):
    passes = (pre_passes + [
        _ReplacePrev(), ssa(strict=False), bool_to_bit(), if_to_phi(_seq_phi)
    ] + post_passes)

    clock_io = gen_clock_io(reset_type, has_enable)

    def seq_inner(cls):
        if reset_type:
            update_register = _UpdateRegister(reset_type,
                                              has_enable, reset_priority)
            cls.__init__ = apply_ast_passes(passes=[update_register],
                                            debug=debug, env=env, path=path,
                                            file_name=file_name)(cls.__init__)

        cls.__call__ = apply_ast_passes(passes, debug=debug, env=env,
                                        path=path,
                                        file_name=file_name)(cls.__call__)
        if annotations is None:
            _annotations = cls.__call__.__annotations__
        else:
            _annotations = annotations

        if "self" in _annotations:
            raise Exception("Assumed self did not have annotation")

        io_args = build_io_args(_annotations)

        class SequentialCircuit(Circuit):
            name = cls.__name__
            io = IO(**io_args) + clock_io
            call_args = [cls()]

            # Monkey patch setattribute for register assign syntax, we could
            # also add this in a Sequential base class, but if we do that we
            # might as well use a metaclass rather than a decorator, but to
            # preserve the old interface we do this for now
            cls.__setattr__ = sequential_setattr

            # also need to patch getattr to support list attributes (see
            # test_sequential2:test_sequential2_pre_unroll)
            cls.__getattribute__ = sequential_getattribute

            call_args += build_call_args(io, _annotations)

            call_result = cls.__call__(*call_args)
            wire_call_result(io, call_result, cls.__call__.__annotations__)
        return SequentialCircuit
    return seq_inner



from collections import defaultdict
import os


__magma_codegen_debug_info = False
if os.environ.get("MAGMA_CODEGEN_DEBUG_INFO", False):
    __magma_codegen_debug_info = True


def set_codegen_debug_info(val):
    global __magma_codegen_debug_info
    __magma_codegen_debug_info = val


def get_codegen_debug_info():
    return __magma_codegen_debug_info


def make_relative(path):
    cwd = os.getcwd()
    common_prefix = os.path.commonprefix([cwd, path])
    return os.path.relpath(path, common_prefix)


class keydefaultdict(defaultdict):
    """See: https://stackoverflow.com/questions/2912231/is-there-a-clever-way-to-pass-the-key-to-defaultdicts-default-factory"""  # noqa
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError( key )  # pragma: no cover
        else:
            ret = self[key] = self.default_factory(key)
            return ret



from .t import In


def MonitorIO(circuit):
    IO = []
    for name, port in circuit.interface.ports.items():
        IO += [name, In(type(port))]
    return IO


def make_monitor_ports(circuit):
    return {name: In(type(port))
            for name, port in circuit.interface.ports.items()}


class MonitorGenerator:
    pass



from magma.array import Array
from magma.circuit import Circuit
from magma.digital import Digital
from magma.ref import (DefnRef, InstRef, NamedRef, ArrayRef, TupleRef,
                       PortViewRef)
from magma.t import Type, Direction
from magma.tuple import Tuple
from magma.view import InstView, PortView


def value_to_verilog_name(value):
    if isinstance(value, Array) and not issubclass(value.T, Digital):
        elems = ", ".join(value_to_verilog_name(t) for t in reversed(value))
        return f"'{{{elems}}}"
    elif isinstance(value, Tuple):
        raise NotImplementedError("Inlining unflattened tuple")
    elif isinstance(value, InstView):
        prefix = ""
        if value.parent is not None:
            prefix = value_to_verilog_name(value.parent) + "."
        return prefix + value.inst.name
    return verilog_name(value.name)


def verilog_name(name, inst_sep="_"):
    if isinstance(name, PortViewRef):
        curr = name.view.parent
        hierarchical_path = curr.inst.name + "."
        while isinstance(curr.parent, InstView):
            hierarchical_path = curr.parent.inst.name + "." + hierarchical_path
            curr = curr.parent
        return hierarchical_path + verilog_name(name.view.port.name)
    if isinstance(name, DefnRef):
        return str(name)
    if isinstance(name, InstRef):
        return f"{name.inst.name}{inst_sep}{str(name)}"
    if isinstance(name, NamedRef):
        return str(name)
    if isinstance(name, ArrayRef):
        array_name = verilog_name(name.array.name)
        if issubclass(name.array.T, Digital):
            index = f"[{name.index}]"
        else:
            index = f"_{name.index}"
        return f"{array_name}{index}"
    if isinstance(name, TupleRef):
        tuple_name = verilog_name(name.tuple.name)
        index = name.index
        if name.root() is None:
            # Not a named temporary that will be flattened to a wire
            try:
                int(index)
                # python/coreir don't allow pure integer names
                index = f"_{index}"
            except ValueError:
                pass
        return f"{tuple_name}_{index}"
    raise NotImplementedError(name, type(name))


def _get_top_level_ref(ref):
    if isinstance(ref, ArrayRef):
        return _get_top_level_ref(ref.array.name)
    if isinstance(ref, TupleRef):
        return _get_top_level_ref(ref.tuple.name)
    return ref


def _sanitize(name):
    return name.replace(".", "_").replace("[", "_").replace("]", "")


def convert_values_to_verilog_str(cls, value):
    if isinstance(value, Type):
        if value.is_input():
            value = value.value()
        if not isinstance(_get_top_level_ref(value.name), DefnRef):
            if value not in cls.inline_verilog_wire_map:
                # Insert a wire so it can't be inlined out
                temp_name = f"_magma_inline_wire"
                temp_name += f"{cls.inline_verilog_wire_counter}"
                with cls.open():
                    temp = type(value).qualify(Direction.Undirected)(
                        name=temp_name
                    )
                    cls.inline_verilog_wire_counter += 1
                    temp @= value
                    temp.unused()
                    cls.inline_verilog_wire_map[value] = temp
            value = cls.inline_verilog_wire_map[value]
        return value_to_verilog_name(value)
    if isinstance(value, PortView):
        if value.port.is_input():
            raise NotImplementedError()
        if value not in cls.inline_verilog_wire_map:
            ref = _get_top_level_ref(value.port.name)
            if isinstance(ref, InstRef):
                defn = ref.inst.defn
            else:
                assert isinstance(ref, DefnRef)
                defn = ref.defn
            with defn.open():
                temp = type(value.port).qualify(Direction.Undirected)(
                    name=f"_magma_inline_wire{cls.inline_verilog_wire_counter}"
                )
                cls.inline_verilog_wire_counter += 1
                temp @= value.port
                temp.unused()
                temp = PortView(temp, value.parent)
                cls.inline_verilog_wire_map[value] = temp
            value = cls.inline_verilog_wire_map[value]
        return value_to_verilog_name(value)
    return str(value)
